/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright © 2005-2011, David Woodhouse
 *
 * David Woodhouse <dwmw2@infradead.org> 
 * Generic app execution and other stuff by Moises Silva <moises.silva@gmail.com> 
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief V.110 application -- accept incoming v.110 call and spawn a custom application
 *
 * \author\verbatim David Woodhouse <dwmw2@infradead.org> \endverbatim
 * 
 * \ingroup applications
 */

/*** MODULEINFO
	<defaultenabled>no</defaultenabled>
 ***/

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision$")

#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/lock.h"
#include "asterisk/logger.h"
#include "asterisk/app.h"
#include "asterisk/transcap.h"

#include <sys/types.h>
#include <sys/wait.h>

/*** DOCUMENTATION
	<application name="V110" language="en_US">
		<synopsis>
			Accepts incoming V.110 data calls and spawns agetty or a custom command in a pseudo-tty.
		</synopsis>
		<syntax>
			<parameter name="command" required="false">
				<para>The name of the command to be run. If unspecified, <filename>/sbin/agetty</filename> will be used with suitable arguments.</para>
			</parameter>
			<parameter name="args..." requires="false">
				<para>Optional arguments to the login command, comma-separated</para>
			</parameter>
		</syntax>
		<description>
			<para>V110 handles incoming data calls. It creates a pseudo-tty and relays data
			from the incoming call to a child process which it spawns on the other end of
			the pseudo-tty.</para>

			<para>If invoked with no arguments, it will spawn a standard 'getty' process
			to handle a local login. If a command is specified, that command is run with
			its stdin, stderr and stdout directed to the pseudo-tty device.</para>

			<para>To dial in to this with a GSM mobile phone, issue the command
			<code>AT+CBST=71,0,1</code> to your phone before the standard <code>ATD</code>
			dial command.</para>
			<para>If used on an incoming call which is not a digital data call, the
			application will return success in order to allow control to pass to the
			next priority in the dialplan. If a call has been handled and has terminated,
			this application will hangup the call before returning</para>
		</description>
	</application>
 ***/


/*
  Testing mode: Send back exactly the same frames we received from the peer. On
  overloaded machines we *still* get the 'UUUUUUUUU' corruption sometimes, which
  seems to happen when Asterisk doesn't keep the TX channel full so it ends up
  being zeroes instead of empty V.110 frames. Would be nice if we could register
  a frame in advance with 'if you have nothing better to send, send this'. 
*/
//#define V110_LOOPBACK

/*
  The generator does work, but makes us more susceptible to 'UUUUUUUUU' corruption.
*/
//#define USE_GENERATOR



static char *app = "V110";

#define URATE_EBITS	0 /* In E-bits or negotiated in-band */
#define URATE_600	1
#define URATE_1200	2
#define URATE_2400	3
#define URATE_3600	4
#define URATE_4800	5
#define URATE_7200	6
#define URATE_8000	7
#define URATE_9600	8
#define URATE_14400	9	/* isdn4linux abuses this for 38400 */
#define URATE_16000	10
#define URATE_19200	11
#define URATE_32000	12

#define URATE_48000	14
#define URATE_56000	15
#define URATE_135	21 /* 134.5 */
#define URATE_100	22
#define URATE_75_1200	23 /* 75 forward, 1200 back */
#define URATE_1200_75	24 /* 1200 forward, 75 back */
#define URATE_50	25
#define URATE_75	26
#define URATE_110	27
#define URATE_150	28
#define URATE_200	29
#define URATE_300	30
#define URATE_12000	31

static char *rates[] = {
	[URATE_600] = "600",
	[URATE_1200] = "1200",
	[URATE_2400] = "2400",
	[URATE_3600] = "3600",
	[URATE_4800] = "4800",
	[URATE_7200] = "7200",
	[URATE_8000] = "8000",
	[URATE_9600] = "9600",
	[URATE_14400] = "14400",
	[URATE_16000] = "16000",
	[URATE_19200] = "19200",
	[URATE_32000] = "32000",
	[URATE_48000] = "48000",
	[URATE_56000] = "56000",
	[URATE_135] = "135",
	[URATE_100] = "100",
	[URATE_75_1200] = "75",
	[URATE_1200_75] = "1200",
	[URATE_50] = "50",
	[URATE_75] = "75",
	[URATE_110] = "110",
	[URATE_150] = "150",
	[URATE_200] = "200",
	[URATE_300] = "300",
	[URATE_12000] = "12000",
};

#define IBUF_LEN 8192
#define OBUF_LEN 1024
#define OBUF_THRESH 16

struct v110_state {
	/* Input v.110 frame buffer */
	unsigned char vframe_in[10];
	unsigned vframe_in_len;

	/* Input data buffer */
	unsigned char ibuf[IBUF_LEN];
	unsigned ibufend;
	unsigned ibufstart;
	unsigned nextibit;

	/* Output data buffer */
	unsigned char obuf[OBUF_LEN];
	unsigned obufend;
	unsigned obufstart;
	unsigned nextobit;

	/* Output v.110 frame buffer */
	unsigned nextoline;
	unsigned last_frame_empty;

	int bufwarning;
	unsigned char cts, rts, sbit;
	int synccount;
	int eofcount;

	/* frame data */
	struct ast_frame f;
	unsigned char friendly[AST_FRIENDLY_OFFSET];
	unsigned char fdata[4096];

	int urate;
	/* hooks to call depending on the rate */
	void (*input_frame)(struct v110_state *, struct ast_frame *);
	void (*fill_outframe)(struct v110_state *, int);	
};

void v110_input_frame_x4(struct v110_state *vs, struct ast_frame *);
void v110_input_frame_x2(struct v110_state *vs, struct ast_frame *);
void v110_input_frame_x1(struct v110_state *vs, struct ast_frame *);
void v110_fill_outframe_x4(struct v110_state *vs, int);
void v110_fill_outframe_x2(struct v110_state *vs, int);
void v110_fill_outframe_x1(struct v110_state *vs, int);

static pid_t loginpty(struct v110_state *vs, struct ast_channel *chan, const char *data, int *ptyfd);

/* Helper for reversing bit-order of data bytes. */
static const unsigned char bit_reverse_table[256] = 
{
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
	0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
	0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
	0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
	0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
	0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
	0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
	0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
	0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
	0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
	0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
	0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
	0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
	0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
	0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
	0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
	0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
};

static int v110_generate(struct ast_channel *chan, void *data, int len, int samples)
{
	struct v110_state *vs = data;
	int want = samples;
	int res = 0;

	while (want > 4000) {
		vs->last_frame_empty = 0;
		vs->fill_outframe(vs, 4000);
		if (ast_write(chan, &vs->f)) {
			res = -1;
			goto out;
		}
		want -= 4000;
	}
	
	/* If the previous Asterisk frame contained only full
	   V.110 frames, and all of them were empty. And if we
	   have no data to send this time either, and want an
	   Asterisk frame of the same size... just re-use the
	   previous one in vs->f. */
	if (vs->nextoline == 10 && !vs->sbit && !vs->nextobit &&
	    vs->obufstart == vs->obufend) {
		if (vs->last_frame_empty && want == vs->f.datalen)
			goto write_frame;

		/* Perhaps next time... */
		vs->last_frame_empty = 1;
	} else {
		/* Not starting at the beginning of a V.110 frame. We
		   definitely can't re-use this Asterisk frame */
		vs->last_frame_empty = 0;
	}
	vs->fill_outframe(vs, want);

 write_frame:
	if (ast_write(chan, &vs->f))
		res = -1;
 out:
	return res;
}

#ifdef V110_LOOPBACK
#undef USE_GENERATOR
#endif

#ifdef USE_GENERATOR
static void *v110_gen_alloc(struct ast_channel *chan, void *data)
{
	return data;
}

static void v110_gen_release(struct ast_channel *chan, void *data)
{
}

static struct ast_generator v110_gen = {
	.alloc = v110_gen_alloc,
	.release = v110_gen_release,
	.generate = v110_generate
};
#endif

static int v110_exec(struct ast_channel *chan, const char *data)
{
	int res = -1;
	int apppid = -1;
	int appstatus = 0;
	int urate = -1;
	int primelen = 0;
	const char *urate_var = NULL;
	struct ast_frame *f = NULL;
	struct v110_state *vs = NULL;
	int ptyfd = -1;

	if (!(chan->transfercapability & AST_TRANS_CAP_DIGITAL)) {
	  ast_verb(3, "Not an async V.110 call; capabilities = [%d]\n", chan->transfercapability);
		/* Return success to allow the dialplan to move on to the next priority */
		return 0;
	}

	/* mISDN puts it into MISDN_URATE variable. Dahdi doesn't seem to tell us at all. */
	urate_var = pbx_builtin_getvar_helper(chan, "MISDN_URATE");
	if (urate_var) {
		urate = atoi(urate_var);
	} else {
		urate_var = pbx_builtin_getvar_helper(chan, "ISDN_V110_URATE");
		if (urate_var)
		urate = atoi(urate_var);
	}
	if (urate == -1) {
		ast_log(LOG_NOTICE, "V.110 assuming 9600 rate, use ISDN_V110_URATE variable to override\n");
		urate = URATE_9600;
	}

	vs = ast_calloc(1, sizeof(*vs)); 
	if (!vs) {
		ast_log(LOG_ERROR, "Allocation of V.110 data structure failed\n");
		res = -1;
		goto out;
	}

	if (!strcasecmp(chan->tech->type, "mISDN")) {
		/* why isn't transfercapability digital enough for mISDN channel driver?? */
		pbx_builtin_setvar_helper(chan, "MISDN_DIGITAL_TRANS", "1");
	} 
	/* we could restrict usage to DAHDI or mISDN channels, but, the truth is that this application only cares about 
	   the provided V.110 bit stream, if some other channel can provide that, good for them */

	/* TODO: Waste bits between characters instead of relying on flow control */
	switch (urate) {
	case URATE_1200:
	case URATE_2400:
	case URATE_4800:
	case URATE_7200:
	case URATE_8000:
	case URATE_9600:
		vs->input_frame = v110_input_frame_x4;
		vs->fill_outframe = v110_fill_outframe_x4;
		primelen = 200;
		break;

	case URATE_12000:
	case URATE_16000:
	case URATE_19200:
		vs->input_frame = v110_input_frame_x2;
		vs->fill_outframe = v110_fill_outframe_x2;
		primelen = 200;
		break;

	case URATE_14400: /* NB. isdn4linux 38400 */
	case URATE_32000:
		vs->input_frame = v110_input_frame_x1;
		vs->fill_outframe = v110_fill_outframe_x1;
		primelen = 200;
		break;

	default:
		ast_log(LOG_NOTICE, "V.110 call at rate %d not supported\n", urate);
		goto out;
	}

	vs->urate = urate;
	vs->nextoline = 10;
	vs->rts = 1;
	vs->sbit = 1;
	vs->cts = 1;
	vs->synccount = 10;
	vs->bufwarning = 5;
	vs->f.data.ptr = vs->fdata;
	vs->f.offset = AST_FRIENDLY_OFFSET;
	vs->f.subclass.codec = chan->readformat;
	vs->f.frametype = AST_FRAME_VOICE;

	apppid = loginpty(vs, chan, data, &ptyfd);
	if (apppid == -1) {
		goto out;
	}

	ast_verb(3, "Accepting V.110 call\n");
	if (chan->_state != AST_STATE_UP) {	
		ast_answer(chan);
	}

	v110_generate(chan, vs, 0, primelen);

#ifdef USE_GENERATOR
	res = ast_activate_generator(chan, &v110_gen, vs);
	if (res) {
		ast_log(LOG_ERROR, "Failed to activate generator on '%s'\n", chan->name);
		goto out;
	}
#endif
	ast_set_priority(1);

	res = 0;
	while (ast_waitfor(chan, -1) > -1) {
		int r, want;

		f = ast_read(chan);
		if (!f) {
			break;
		}

		if (f->frametype != AST_FRAME_VOICE) {
			ast_frfree(f);
			continue;
		}

		f->delivery.tv_sec = 0;
		f->delivery.tv_usec = 0;

		vs->input_frame(vs, f);

#ifndef V110_LOOPBACK
		/* normal operation */
#ifndef USE_GENERATOR
		if (v110_generate(chan, vs, 0, f->datalen)) {
			res = -1;
			break;
		}
#endif /* GENERATOR */

#else
		/* loopback mode for testing */
		if (0 && f->datalen >= 8)
			ast_log(LOG_NOTICE, "Received frame: %02x %02x %02x %02x %02x %02x %02x %02x\n",
				((unsigned char *)f->data.ptr)[0],
				((unsigned char *)f->data.ptr)[1],
				((unsigned char *)f->data.ptr)[2],
				((unsigned char *)f->data.ptr)[3],
				((unsigned char *)f->data.ptr)[4],
				((unsigned char *)f->data.ptr)[5],
				((unsigned char *)f->data.ptr)[6],
				((unsigned char *)f->data.ptr)[7]);
		if (ast_write(chan, f)) {
			res = -1;
			break;
		}
#endif

		/* Flush v.110 incoming buffer */
		if (vs->ibufend > vs->ibufstart) {
			want = vs->ibufend - vs->ibufstart;
		} else if (vs->ibufend < vs->ibufstart) {
			want = IBUF_LEN - vs->ibufstart;
		} else {
			want = 0;
		}
		if (want) {
			r = write(ptyfd, &vs->ibuf[vs->ibufstart], want);
			/* Ignore EPIPE too; we'll detect EOF on input */
			if (r < 0 && (errno == EAGAIN || errno == EPIPE)) {
				r = 0;
			}
			if (r < 0) {
				ast_log(LOG_DEBUG, "Error writing to pty: %s\n", strerror(errno));
				res = -1;
				goto out;
			}
			vs->ibufstart += r;
			if (vs->ibufstart == IBUF_LEN) {
				vs->ibufstart = 0;
			}

			/* Set flow control state. */
			if (r < want) {
				vs->rts = 1;
			} else {
				vs->rts = 0;
			}
		}

		/* Replenish v.110 outgoing buffer */
		if (vs->obufend >= vs->obufstart) {
			if (vs->obufend - vs->obufstart < OBUF_THRESH) {
				want = OBUF_LEN - vs->obufend - !vs->obufstart;
			} else {
				want = 0;
			}
		} else {
			if (vs->obufstart + OBUF_LEN - vs->obufend < OBUF_THRESH) {
				want = vs->obufstart - vs->obufend - 1;
			} else {
				want = 0;
			}
		}
		if (want) {
			r = read(ptyfd, &vs->obuf[vs->obufend], want);
			if (r < 0) {
				/* Don't abort on read error, until we've flushed data */
				if (errno == EAGAIN || vs->obufstart != vs->obufend) {
					r = 0;
				} else {
					/* It's expected that we get error when the other end closes the pipe or pty */
					ast_log(LOG_DEBUG, "Error reading from pty: %s\n", strerror(errno));
					res = -1;
					goto out;
				}
			}
			vs->obufend += r;
			if (vs->obufend == OBUF_LEN) {
				vs->obufend = 0;
			}
		}
	}
out:
	ast_log(LOG_DEBUG, "Exiting app_v110\n");

#ifdef USE_GENERATOR
	ast_deactivate_generator(chan);
#endif

	vs->sbit = 1;
	v110_generate(chan, vs, 0, primelen);

	/* free structures */
	if (vs) {
		ast_free(vs);
	}
	if (f) {
		ast_frfree(f);
	}

	ast_softhangup (chan, AST_SOFTHANGUP_SHUTDOWN);

	/* cleanup custom app */
	if (apppid > -1) {
		close(ptyfd);
		ast_log(LOG_DEBUG, "Reaping child pid %d\n", apppid);
		kill(apppid, SIGTERM);
		waitpid(apppid, &appstatus, 0);
		ast_log(LOG_DEBUG, "Child status = %d\n", WEXITSTATUS(appstatus));
	} 

	return res;
}

static void v110_process_frame(struct v110_state *vs) 
{
	int octet;

	if (0) ast_log(LOG_NOTICE, "frame %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
		vs->vframe_in[0], vs->vframe_in[1], vs->vframe_in[2],
		vs->vframe_in[3], vs->vframe_in[4], vs->vframe_in[5],
		vs->vframe_in[6], vs->vframe_in[7], vs->vframe_in[8], 
		vs->vframe_in[9]);

	/* Check that line 5 (E-bits) starts '1011' (reversed). */
	if ((vs->vframe_in[5] & 0xf0) != 0xb0) 
		return;

	/* Check that each other octet starts with '1' */
	if (!(vs->vframe_in[1] & vs->vframe_in[2] & vs->vframe_in[3] & 
	      vs->vframe_in[4] & vs->vframe_in[6] & vs->vframe_in[7] & 
	      vs->vframe_in[8] & vs->vframe_in[9] & 0x80))
		return;
		
	/* Extract flow control signal from last octet */
	if (vs->synccount) {
		if (!--vs->synccount) {
			ast_verb(3, "V.110 synchronisation achieved\n");
			vs->sbit = 0;
			vs->rts = 0;
		}
	} else
		vs->cts = vs->vframe_in[7] & 1;

	for (octet = 1; octet < 10; octet++) {
		unsigned char tmp;

		/* Skip E-bits in line 5 */
		if (octet == 5)
			continue;

		tmp = bit_reverse_table[vs->vframe_in[octet]] & 0x7e;

		/* Search for start bit if not yet found */
		if (!vs->nextibit) {

			/* First check for no zero bits. This will be common */
			if (tmp == 0x7e)
				continue;

			/* Check for start bit being last in the octet */
			if (tmp == 0x3e) {
				vs->nextibit = 1; /* Expecting first data bit now */
				vs->ibuf[vs->ibufend] = 0;
				continue;
			}
			
			/* Scan for the start bit, copy the data bits (of which
			   there will be at least one) into the next byte of ibuf */
			vs->nextibit = 7;
			do {
				tmp >>= 1;
				vs->nextibit--;
			} while (tmp & 1);

			/* Start bit is now (host's) LSB */
			vs->ibuf[vs->ibufend] = tmp >> 1;
			continue;
		}

		tmp >>= 1;

		if (vs->nextibit < 9) {
			/* Add next bits of incoming byte to ibuf */
			vs->ibuf[vs->ibufend] |= tmp << (vs->nextibit-1);

			
			if (vs->nextibit <= 3) {
				/* Haven't finished this byte (including stop) yet */
				vs->nextibit += 6;
				continue;
			}

			tmp >>= (9 - vs->nextibit);
		}

		/* Check for stop bit */
		if (tmp & 1) {
			unsigned newend = (vs->ibufend + 1) & (IBUF_LEN-1);

			if (newend == vs->ibufstart) {
				/* Buffer full. This shouldn't happen because we should
				   have asserted flow control long ago */
				if (vs->bufwarning) {
					vs->bufwarning--;
					ast_log(LOG_NOTICE, "incoming buffer full\n");
				}
				continue;
			} else
				vs->ibufend = newend;
		} else {
			ast_log(LOG_NOTICE, "No stop bit\n");
		}
		
		/* Now, scan for next start bit */
		tmp >>= 1;
		vs->nextibit -= 4;
		while (vs->nextibit && (tmp & 1)) {
			tmp >>= 1;
			vs->nextibit--;
		}
		if (vs->nextibit > 1)
			vs->ibuf[vs->ibufend] = tmp >> 1;
			
	}

}

/* We don't handle multiple multiplexed channels. Nobody really does */
void v110_input_frame_x4(struct v110_state *vs, struct ast_frame *f)
{
	int datalen = f->datalen;
	unsigned char *frame_data = f->data.ptr;

	while (datalen) {
		if (vs->vframe_in_len < 4) {
			/* Find zero octet in buffer */
			if ( (*frame_data) & 0xc0 ) {
				vs->vframe_in_len = 0;
				frame_data++;
				datalen--;
				continue;
			}
			/* Found a suitable byte. Add it. */
			if (++vs->vframe_in_len == 4)
				memset(vs->vframe_in, 0, 10);
			frame_data++;
			datalen--;
			continue;
		}
		/* Add in these two bits */
		vs->vframe_in[vs->vframe_in_len/4] |= 
			((*frame_data) & 0xc0) >> ((vs->vframe_in_len & 3) * 2);

		vs->vframe_in_len++;
		frame_data++;
		datalen--;

		if (vs->vframe_in_len == 40) {
			v110_process_frame(vs);
			vs->vframe_in_len = 0;
		}
	}
}

void v110_input_frame_x2(struct v110_state *vs, struct ast_frame *f)
{
	int datalen = f->datalen;
	unsigned char *frame_data = f->data.ptr;

	while (datalen) {
		if (vs->vframe_in_len < 2) {
			/* Find zero octet in buffer */
			if ( (*frame_data) & 0xf0 ) {
				vs->vframe_in_len = 0;
				frame_data++;
				datalen--;
				continue;
			}
			/* Found a suitable byte. Add it. */
			if (++vs->vframe_in_len == 2)
				memset(vs->vframe_in, 0, 10);
			frame_data++;
			datalen--;
			continue;
		}
		/* Add in these four bits */
		vs->vframe_in[vs->vframe_in_len/2] |= 
			((*frame_data) & 0xf0) >> ((vs->vframe_in_len & 1) * 4);

		vs->vframe_in_len++;
		frame_data++;
		datalen--;

		if (vs->vframe_in_len == 20) {
			v110_process_frame(vs);
			vs->vframe_in_len = 0;
		}
	}
}

void v110_input_frame_x1(struct v110_state *vs, struct ast_frame *f)
{
	int datalen = f->datalen;
	unsigned char *frame_data = f->data.ptr;

	while (datalen) {
		if (!vs->vframe_in_len) {
			/* Find zero octet in buffer */
			if ( (*frame_data)) {
				vs->vframe_in_len = 0;
				frame_data++;
				datalen--;
				continue;
			}
			/* Found a suitable byte. Add it. */
			vs->vframe_in_len++;
			memset(vs->vframe_in, 0, 10);
			frame_data++;
			datalen--;
			continue;
		}
		/* Add byte to frame */
		vs->vframe_in[vs->vframe_in_len] = *frame_data;

		vs->vframe_in_len++;
		frame_data++;
		datalen--;

		if (vs->vframe_in_len == 10) {
			v110_process_frame(vs);
			vs->vframe_in_len = 0;
		}
	}
}

/* Some bitmasks to ease calculation. */
static unsigned char helper1[] = { 0x81, 0x81, 0x81, 0xc1, 0xe1, 0xf1, 0xf9, 0xfd, 0xff };
static unsigned char helper2[] = { 0x81, 0x83, 0x87, 0x8f, 0x9f, 0xbf };

static unsigned char v110_getline(struct v110_state *vs)
{
	unsigned char octet;
	unsigned char tmp;
	int line = vs->nextoline++;
	int place = 2;

	if (line == 10) {
		vs->nextoline = 1;
		return 0x00; /* Header */
	} else if (line == 5) {
		return 0xbf; /* E-bits. 10111111 */
	} else if (line == 2 || line == 7) {
		octet = 0xfe | vs->rts;
	} else {
		octet = 0xfe | vs->sbit;
	}

	/* If we're already sending a byte, finish it */
	if (vs->nextobit) {
		vs->last_frame_empty = 0;

		/* Shift the data byte so that the bit we want is in bit 1 */
		tmp = vs->obuf[vs->obufstart] >> (vs->nextobit - 2);

		/* Mask in the bits we don't want to touch and the stop bit */
		tmp |= helper1[vs->nextobit - 1];

		/* Clear bits in the generated octet to match */
		octet &= bit_reverse_table[tmp];

		if (vs->nextobit < 4) {
			/* There's some of this byte left; possibly just the stop bit */
			vs->nextobit += 6;
			return octet;
		}

		/* We've finished this byte */
		vs->obufstart++;
		if (vs->obufstart == OBUF_LEN)
			vs->obufstart = 0;

		if (vs->nextobit < 5) {
			/* But there's still no room in this octet for any more */
			vs->nextobit = 0;
			return octet;
		}
		/* Work out where to put the next data byte */
		place = 12 - vs->nextobit;
		vs->nextobit = 0;
	} else {
		/* Nothing to follow; start bit of new byte at bit 1 */
		place = 2;
	}

	/* Honour flow control when starting new characters */
	if (vs->cts || vs->obufstart == vs->obufend)
		return octet;

	vs->last_frame_empty = 0;
	/* 'place' is the location within the octet to start the new
	   data byte. It'll be 2 unless we've already got the tail of
	   a previous data byte in this octet. If you're starting at it
	   and think there's an off-by-one error, remember the start bit
	   which is zero, and in bit (place-1). */
	tmp = vs->obuf[vs->obufstart] << place;

	/* The bits we don't want to touch */
	tmp |= helper2[place - 2];

	octet &= bit_reverse_table[tmp];
	vs->nextobit = 8 - place;

	return octet;
}

void v110_fill_outframe_x4(struct v110_state *vs, int datalen)
{
	unsigned char *pos = vs->f.data.ptr;

	if (datalen & 3) {
		vs->last_frame_empty = 0;
		datalen = (datalen + 3) & ~3;
	}

	vs->f.datalen = vs->f.samples = datalen;

	while (datalen) {
		unsigned char tmp = v110_getline(vs);
		pos[0] = 0x3f | (tmp & 0xc0);
		tmp <<= 2;
		pos[1] = 0x3f | (tmp & 0xc0);
		tmp <<= 2;
		pos[2] = 0x3f | (tmp & 0xc0);
		tmp <<= 2;
		pos[3] = 0x3f | tmp;
		pos += 4;
		datalen -= 4;
	}
}

void v110_fill_outframe_x2(struct v110_state *vs, int datalen)
{
	unsigned char *pos = vs->f.data.ptr;

	if (datalen & 1) {
		vs->last_frame_empty = 0;
		vs->f.datalen = datalen = datalen + 1;
	}

	vs->f.datalen = vs->f.samples = datalen;

	while (datalen) {
		unsigned char tmp = v110_getline(vs);
		pos[0] = 0x0f | (tmp & 0xf0);
		tmp <<= 4;
		pos[1] = 0x0f | tmp;
		pos += 2;
		datalen -= 2;
	}
}

void v110_fill_outframe_x1(struct v110_state *vs, int datalen)
{
	unsigned char *pos = vs->f.data.ptr;

	vs->f.datalen = vs->f.samples = datalen;

	while (datalen) {
		*pos = v110_getline(vs);
		pos++;
		datalen--;
	}
}

pid_t loginpty(struct v110_state *vs, struct ast_channel *chan, const char *data, int *ptyfd)
{
	int master;
	pid_t pid = -1; 
	char name[256];
	char *source = "<unknown>";
	int flags;
	char **child_argv = NULL;
	int child_argc = 0;

	if (data && *data) {
		char *arg, *userargs = ast_strdupa(data);

		while ( (arg = strsep(&userargs, ",")) && *arg) {
			child_argv = realloc(child_argv, sizeof(char *) * (child_argc + 2));
			if (!child_argv) {
				ast_log(LOG_ERROR, "Failed to allocate child arguments\n");
				goto out;
			}
			child_argv[child_argc++] = arg;
		}
		child_argv[child_argc] = NULL;
	} else {
		child_argv = malloc(sizeof(char *) * 8);
		if (!child_argv) {
			ast_log(LOG_ERROR, "Failed to allocate child arguments\n");
			goto out;
		}
		child_argv[child_argc++] = "/sbin/agetty";
		if (chan->caller.id.number.valid) {
			child_argv[child_argc++] = "-H";
			source = child_argv[child_argc++] = chan->caller.id.number.str;
		}
		child_argv[child_argc++] = rates[vs->urate];
		child_argv[child_argc++] = name;
		child_argv[child_argc++] = "vt100";
		child_argv[child_argc++] = NULL;
	}

	master = getpt();

	if (master < 0) {
		ast_log (LOG_NOTICE, "Failed to allocate pty: %s\n",
			 strerror(errno));
		goto out;
	}

	if (grantpt(master)) {
		ast_log(LOG_NOTICE, "grantpt() failed: %s\n", strerror(errno));
		close(master);
		goto out;
	}

	if (unlockpt(master)) {
		ast_log(LOG_NOTICE, "unlockpt() failed: %s\n", strerror(errno));
		close(master);
		goto out;
	}

	flags = fcntl(master, F_GETFL);
	fcntl(master, F_SETFL, flags | O_NONBLOCK);

	if (ptsname_r(master, name, sizeof(name))) {
		ast_log(LOG_NOTICE, "ptsname_r() failed: %s\n", strerror(errno));
		close(master);
		goto out;
        }

	if ((pid = ast_safe_fork(1)) < 0) {
		ast_log(LOG_WARNING, "Failed to fork() to launch login process %s\n", strerror(errno));
		close (master);
		goto out;
	}

	if (!pid) {
		ast_set_priority(0);

		/* Close everything */
		close(0);
		ast_close_fds_above_n(0);
		setsid();

		setenv ("TERM", "vt100", 1);
		setenv ("V110_RATE", rates[vs->urate], 1);
		setenv ("V110_SOURCE", source, 1);
		setenv ("V110_PTY", name, 1);

		if (data && *data) {
			/* If a command was specified, open the slave PTY for it */
			int slave = open(name, O_RDWR);
			if (slave < 0)
				_exit(1);

			dup2(0, 1);
			dup2(0, 2);
		}
			
		execv(child_argv[0], child_argv);

		fprintf(stderr, "Failed to spawn child process: %s\n", strerror(errno));
		_exit(1);
	}
	ast_verb(3, "Launched login process\n");
	*ptyfd = master;
   out:
	free (child_argv);
	return pid;
}

static int load_module(void)
{
	return ast_register_application_xml(app, v110_exec) ? AST_MODULE_LOAD_DECLINE : AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	return ast_unregister_application(app);
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "V.110 dialin");

